.\"
.\" troff -ms % | lpr
.\"
.\" revision date - change whenever this file is edited
.ds RD 26 September 1991
.nr PO 1.2i	\" page offset 1.2 inches
.nr PD .7v	\" inter-paragraph distance
.\"
.EH 'Token Scanning Library'- % -''
.OH ''- % -'Token Scanning Library'
.OF 'Revision date:\0\0\*(RD''Printed:\0\0\n(dy \*(MO 19\n(yr'
.EF 'Revision date:\0\0\*(RD''Printed:\0\0\n(dy \*(MO 19\n(yr'
.\"
.\" I - italic font (taken from -ms and changed)
.de I
.nr PQ \\n(.f
.if t \&\\$3\\f2\\$1\\fP\&\\$2
.if n .if \\n(.$=1 \&\\$1
.if n .if \\n(.$>1 \&\\$1\c
.if n .if \\n(.$>1 \&\\$2
..
.de IS	\" interface routine description header start
.DS L
.ta .8i
.ft B
..
.de IE	\" interface routine description header end
.DE
.ft R
..
.TL
TS \*- A Simple Token Scanning Library
.sp .5v
.ps 10
Release 1.02
.AU
Paul DuBois
dubois@primate.wisc.edu
.AI
Wisconsin Regional Primate Research Center
Revision date:\0\0\*(RD
.LP
Applications often wish to pull apart strings into individual tokens.
This document describes TS, a library consisting of
an unsophisticated set of routines providing simple token scanning operations.
.LP
String tokenizing can often be done satisfactorily using
.I strtok()
or equivalent function from the C library.
When such routines are insufficient, the routines described here may be useful.
They offer, for example, quote and escape character parsing, and
configurability of underlying scanning properties, within the confines
of a fixed interface.
TS provides a simple built-in scanner, which
may be replaced by alternate routines as
desired.
Applications may switch back and forth between scanners
on the fly.
.NH
Installation
.LP
This release of TS is configured using
.I imake
and the standard WRPRC multiple-project configuration files, so you
also need to obtain the WRPRC configuration distribution (version 1.03 or
higher) if you want to build it the usual way.
(If you want to avoid
.I imake ,
the
.I Makefile
is simple enough that you should be able to tweak it by hand.)
.LP
There is one library to be built,
.I libtokenscan.a .
That library should be installed in a system library directory.
The header file
.I tokenscan.h
should be installed in a header file directory.
.NH
Example
.LP
The canonical method of tokenizing a string with TS is as follows:
.LP
.DS
char	buf[size], *p;

	\fI...set contents of buf here...\fR
	TSScanInit (buf);	/* initialize scanner */
	while ((p = TSScan ()) != (char *) NULL)
	{
		\fI...do something here with token pointed to by p here...\fR
	}
.DE
.LP
The scanner is initialized by passing the string to be scanned to
.I TSScanInit()
and
.I TSScan()
is called to get pointers to successive tokens.
.I TSScan()
returns NULL when there are no more.
.NH
Behavior of the Default Scanner
.LP
The default scanner assumes that strings are terminated with a null
byte, and is destructive in that it modifies the string scanned; it
writes nulls at the end of each token found.
.LP
The default delimiter, quote and escape characters are given below.
.LP
.TS
center tab(:);
l l .
delimiter:space tab
quote:" '
escape:\e
.TE
.LP
In the simplest case, tokens are sequences of characters between
delimiters.
Since the default delimiters are the whitespace characters space and
tab, tokens are sequences of non-whitespace characters.
(Newlines at the end of strings are ignored.)
.LP
.DS
.ta 1.5i 2i
This is a line	\(->	<This> <is> <a> <line>
.DE
.LP
Quotes may be used to include whitespace within a token.
Quotes must match; hence one quote character may be used to quote
another kind of quote character, if there is more than one.
.LP
.DS
.ta 1.5i 2i
"This is" a line	\(->	<This is> <a> <line>
This" "is a line	\(->	<This is> <a> <line>
"'" '"'	\(->	<'> <">
.DE
.LP
The escape character turns off any special meaning of the next
character, including another escape character.
.LP
.DS
.ta 1.5i 2i
What\e's up	\(->	<What's> <up>
\e\e is the escape	\(->	<\e> <is> <the> <escape>
.DE
.LP
You can replace the delimiter, quote or escape character sets.
This will change the particular characters that trigger the above
behaviors, without changing the way the default scan routine works.
Or you can change the scan routine to make the scanner behave in
entirely different ways.
.NH
Programming Interface
.LP
Source files using TS routines should include
.I tokenscan.h .
Programs using TS should be linked with
.I "\-ltokenscan" .
.LP
A scanner is described by a data structure:
.LP
.DS
typedef struct TSScanner TSScanner;
struct TSScanner
{
	void	(*scanInit) ();
	char	*(*scanScan) ();
	char	*scanDelim;
	char	*scanQuote;
	char	*scanEscape;
}
.DE
.LP
Scanner structures may be obtained or installed with
.I TSSetScanner()
and
.I TSGetScanner() .
.LP
For each string to be scanned, the application passes a pointer to it to
.I TSScanInit() ,
which takes care of scan initialization.
If the application requires initialization to be performed in addition to
that done internally by TS, a pointer to a routine that does so should
be installed in the
.I scanInit
field of the scanner data structure.
It takes one argument, a pointer to the string to be scanned.
The default
.I scanInit
is NULL, which does nothing.
.LP
.I scanDelim ,
.I scanQuote
and
.I scanEscape
are pointers to a null-terminated strings consisting of
the set of characters to be considered token delimiters, quote
characters and escape characters, respectively.
The default values have already been described above.
.LP
.I scanScan
points to the
routine that does the actual scanning.
It is called by
.I TSScan()
and should be declared to take no arguments and return a character
pointer to the next token in the current scan buffer.
Normally, this routine does the following:
call
.I TSGetScanPos()
to get the current scan position, scan the token, call
.I TSSetScanPos()
to update the scan position, then return a pointer to the beginning of
the token.
If there are no more tokens in the scan buffer, the routine should return
NULL, and should continue to do so until
.I TSScanInit()
is called again.
.LP
The public routines in the TS library are described below.
.IS
void TSScanInit (p)
char	*p;
.IE
Initializes the scanning routines to make
the character string pointed to by
.I p
the current scan buffer.
The default scanner is destructive (inserts null bytes after each
token, strips quotes and backslashes),
so make a copy of the string if you need to maintain an intact version.
.IS
char *TSScan ()
.IE
Returns a pointer to the next token in the current scan buffer, NULL if
there are no more.
The token is terminated by a null byte.
Scan behavior may be modified by substituting alternate scan routines.
.LP
Once
.I TSScan()
returns NULL, it continues to do so
until the scanner is reinitialized with another call to
.I TSScanInit() .
.IS
void TSGetScanner (p)
TSScanner	*p;
.IE
Gets the current scanner information (initialization and scan procedures, and
delimiter, quote and escape character sets) into the structure pointed to by
.I p .
.IS
void TSSetScanner (p)
TSScanner	*p;
.IE
Installs a scanner.
Any field in
.I p
with a NULL value causes the default value from the built-in scanner
to be reinstalled.
If
.I p
itself if NULL, all built-in scanner values are reinstalled.
.IS
void TSGetScanPos (p)
char	**p;
.IE
Puts the current position within the current scan buffer into the
argument, which should be passed as the address of a character pointer.
This is useful when you want to scan only enough of the buffer to partially
classify it, then use the rest in some other way.
.IS
void TSSetScanPos (p)
char	*p;
.IE
Set the current scan position to
.I p .
.IS
int TSIsScanDelim (c)
char	*c;
.IE
Returns non-zero if
.I c
is a member of the current delimiter character set.
.IS
int TSIsScanQuote (c)
char	*c;
.IE
Returns non-zero if
.I c
is a member of the current quote character set.
.IS
int TSIsScanEscape (c)
char	*c;
.IE
Returns non-zero if
.I c
is a member of the current escape character set.
.NH 2
Overriding Scanning Routines
.LP
It is possible to switch back and forth between scan procedures on the
fly, even within the current scan buffer.
The general procedure is to use
.I TSGetScanner()
to get the current scanner information,
and
.I TSSetScanner()
to install a new one and reinstall the old one when done with
the new.
If you switch between more than two scanners, another method may be
necessary.
.LP
It is possible to modify the default scanner without replacing it.
For instance, you could change the default delimiter set but leave everything
else the same,
as follows:
.LP
.DS
TSScanner scanStruct;

TSGetScanner (&scanStruct);
scanStruct.scanDelim = " \et:;?,!";
TSSetScanner (&scanStruct);
.DE
.NH
Miscellaneous
.LP
A scanner can be nondestructive with respect to the line being scanned
by using a scan routine that copies characters out of the scanned
line into a second buffer and returning a pointer to the second buffer.
The second buffer must be large enough to hold the largest possible token,
of course.
If the second buffer is a fixed area, the host application must be careful
not to call
.I TSScan()
again until it is done with the current token, or else make a copy of it
first.
If the second buffer is dynamically allocated, the application must
be ready to do storage management on the tokens returned.
.LP
Some scanners might not use delimiter, quote or escape characters at all,
particularly if token boundaries are context sensitive.
.NH
Distribution and Update Availability
.LP
The TS
distribution may be freely circulated and is available for
anonymous FTP access under the
.I ~ftp/pub
directory on host
.I ftp.primate.wisc.edu .
Updates appear there as they become available.
.LP
The WRPRC configuration distribution is available on
.I ftp.primate.wisc.edu
as well.
.LP
If you do not have FTP access, send requests to the address
.I software@primate.wisc.edu .
Bug reports, questions, suggestions and comments may be sent to
this address as well.
